from typing import List, Dict
import sympy as sp


def criticalPoints(function: sp.Function, x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the critical points of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: critical points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import criticalPoints
    >>> x = symbols('x')
    >>> criticalPoints(x**2, x)
    [0]
    """
    return sp.solve(sp.diff(function, x), x, dict=True)

def minimas(function: sp.Function, x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the minimas of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: minimas of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import minimas
    >>> x = symbols('x')
    >>> minimas(x**2, x)
    [0]
    """
    return [point for point in criticalPoints(function, x) if sp.diff(function, x, 2).subs(point) > 0]

def maximas(function: sp.Function, x: sp.Symbol = sp.Symbol("x")) -> List[sp.Number]:
    """Returns the maximas of a function of one variable.
    :param function: function of one variable
    :param x: variable
    :return: maximas of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import maximas
    >>> x = symbols('x')
    >>> maximas(-x**2, x)
    [0]
    """
    return [point for point in criticalPoints(function, x) if sp.diff(function, x, 2).subs(point) < 0]



import sympy as sp
import numpy as np


def TnValue(f: sp.Function,
            x0: float,
            n: int,
            symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor value of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor value
    :param symbol: symbol of x
    :return: nth Taylor value of f(x) at x0
    """
    df = sp.diff(f, symbol, n).subs(symbol, x0)
    return df / np.math.factorial(n) * (symbol - x0)**n

def Tn(f: sp.Function,
       x0: float,
       n: int,
       symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor polynomial of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor polynomial
    :param symbol: symbol of x
    :return: nth Taylor polynomial of f(x) at x0
    """
    tn = 0
    for i in range(n + 1):
        tn += TnValue(f, x0, i, symbol)
    return tn

def TnErrorFunction(f: sp.Function,
                    x0: float,
                    n: int,
                    symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the nth Taylor error function of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param n: nth Taylor error
    :param symbol: symbol of x
    :return: nth Taylor error function of f(x) at x0
    """
    return (f - Tn(f, x0, n, symbol))

def TnError(f: sp.Function,
            x0: float,
            a: float,
            b: float,
            n: int,
            symbol: sp.Symbol = sp.Symbol("x")) -> float:
    """
    Returns the nth Taylor error of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param n: nth Taylor error
    :param symbol: symbol of x
    :return: nth Taylor error of f(x) at x0
    """
    errf = TnErrorFunction(f, x0, n, symbol)
    xx = np.linspace(a, b, 100)
    errfnp = sp.lambdify(symbol, errf, "numpy")
    yy = errfnp(xx)
    return np.max(np.abs(yy))

def LagrangeRemainder(f: sp.Function,
                      x0: float,
                      a: float,
                      b: float,
                      n: int,
                      symbol: sp.Symbol = sp.Symbol("x")) -> sp.Function:
    """
    Returns the Lagrange remainder of f(x) at x0.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param n: nth Lagrange remainder
    :param symbol: symbol of x
    :return: nth Lagrange remainder of f(x) at x0
    """
    n += 1

    x = sp.Symbol("x")
    M = sp.maximum((x - x0)**n, x, sp.Interval(abs(a), abs(b)))
    m = sp.root(M, n) + x0

    dfn = sp.diff(f, symbol, n)

    if dfn.is_negative:
        dfn = -dfn

    g = dfn * M / np.math.factorial(n)

    T = sp.maximum(g, symbol, sp.Interval(x0, m))

    if float(T) == 0:
        if n > 2:
            print(f"Warning: Lagrange Remainder is likely incorrect (T = 0) - returning for n={n - 2} instead") # pylint: disable=line-too-long
            return LagrangeRemainder(f, x0, a, b, n - 2, symbol)
        else:
            print("Warning: Lagrange Remainder is likely incorrect (T = 0) - returning 0")
            return 0
    return sp.Abs(T)

def TnMin(f: sp.Function,
          x0: float,
          a: float,
          b: float,
          eps: float,
          symbol: sp.Symbol = sp.Symbol("x")) -> int:
    """
    Returns the minimum n such that the nth Taylor error of f(x) at x0 is less than eps.
    :param f: function f(x)
    :param x0: point x0
    :param a: left bound
    :param b: right bound
    :param eps: maximum error
    :param symbol: symbol of x
    :return: minimum n such that the nth Taylor error of f(x) at x0 is less than eps
    """
    n = 0
    while TnError(f, x0, a, b, n, symbol) > eps:
        n += 1
    return n



from __future__ import annotations

from typing import Optional, Union

import sympy as sp
from sympy import Rational


class Convergence:
    def __init__(self,
                 does: bool = False,
                 to: Optional[Union[float, Rational]] = None) -> None:
        self._does = does
        self._to = to

    @property
    def does(self) -> bool:
        return self._does

    @property
    def doesNot(self) -> bool:
        return not self._does

    @property
    def to(self) -> Union[float, Rational]:
        return self._to

    @property
    def toRational(self) -> Rational:
        return Rational(self._to)

    @property
    def toFloat(self) -> float:
        return float(self.to)

    def __str__(self) -> str:
        if self.does:
            return f'converges to {self.to} (~{self.toFloat}))'
        return 'diverges'

    @staticmethod
    def convergence(to: float) -> Convergence:
        return Convergence(True, to)

    @staticmethod
    def divergence() -> Convergence:
        return Convergence(False, None)


def regulaFalsi(f: sp.Function,
                a: float,
                b: float,
                eps: float,
                maxiter: int,
                symbol: sp.Symbol = sp.Symbol("x")) -> Optional[Convergence]:
    """
    Regula Falsi method for finding a root of f(x) = 0 in the interval [a, b].
    :param f: function f(x) = 0
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param eps: tolerance
    :param maxiter: maximum number of iterations
    :param symbol: symbol of the function (defaults to x)
    :return: (flag, iter)

    Examples
    ========
    >>> import sympy as sp
    >>> from anlis.convergence import regulaFalsi
    >>> x = sp.Symbol("x")
    >>> regulaFalsi(x ** 2 - 5, 0.5, 2, 1e-6, 1000)
    (regula falsi) converged after 6 iterations
    Converges to 2.2360678206352302 (~2.2360678206352302)
    """
    g = sp.lambdify(symbol, f, "numpy")
    i = 1
    flag = True
    while flag:
        x = a - g(a)*(b-a) / (g(b) - g(a))
        # print ('Schnittpunkt bei 9615.12e•
        if g(a)*g(x) < 0:
            b = x # Vorzeichenwechsel in [a, x J
        else:
            a = x # Vorzeichenwechsel in [x, b]
            # print( ' Vorzeichenwechsel rechts' )
            i += 1
            flag = abs(g(x)) > eps and i < maxiter
    if i < maxiter:
        print("(regula falsi) converged after %d iterations" % i)
        return Convergence.convergence(x)
    else:
        print("(regula falsi) didn't converge after %d iterations" % maxiter)
        return None

def newton(f: sp.Function,
           x0: float,
           eps: float,
           maxiter: int,
           symbol: sp.Symbol = sp.Symbol("x")) -> Optional[Convergence]:
    """
    Newton method for finding a root of f(x) = 0.
    :param f: function f(x) = 0
    :param x0: initial guess
    :param eps: tolerance
    :param maxiter: maximum number of iterations
    :param symbol: symbol of the function (defaults to x)
    :return: (flag, iter)

    Examples
    ========
    >>> import sympy as sp
    >>> from anlis.convergence import newton
    >>> x = sp.Symbol("x")
    >>> newton(x ** 2 - 5, 2, 1e-6, 1000)
    (newton) converged after 4 iterations
    converges to 51841/23184 (~2.236067977915804))
    """
    i = 1
    flag = True
    while flag:
        x = x0 - f.subs(symbol, x0) / f.diff(symbol).subs(symbol, x0)
        i += 1
        flag = abs(f.subs(symbol, x)) > eps and i < maxiter
        x0 = x
    if i < maxiter:
        print(f"(newton) converged after {i} iterations")
        return Convergence.convergence(x)
    else:
        print(f"(newton) didn't converge after {maxiter} iterations")
        return None

def ratioTest(ak: sp.Function, symbol: sp.Symbol = sp.Symbol("k")) -> Optional[Convergence]:
    """
    Ratio test for convergence of a series.
    :param ak: a_k
    :param symbol: symbol of the series
    :return: convergence

    Examples
    ========

    >>> import sympy as sp
    >>> from anlis.convergence import ratioTest
    >>> k = sp.Symbol("k")
    >>> ratioTest(k / 5 ** k, k)
    Converges to 1/5
    """
    ak1 = ak.subs(symbol, symbol + 1)
    to = sp.limit(sp.Abs(ak1 / ak), symbol, sp.oo)

    if to == 1:
        return None
    if to > 1:
        return Convergence.divergence()
    return Convergence.convergence(to)

def rootTest(ak: sp.Function, symbol: sp.Symbol = sp.Symbol("k")) -> Optional[Convergence]:
    """
    Root test for convergence of a series.
    :param ak: a_k
    :param symbol: symbol of the series
    :return: convergence

    Examples
    ========
    >>> import sympy as sp
    >>> from anlis.convergence import rootTest
    >>> k = sp.Symbol("k")
    >>> rootTest(k / 5 ** k, k)
    Converges to 1/5
    """
    f = sp.root(sp.Abs(ak), symbol)
    to = sp.limit(f, symbol, sp.oo)

    if to == 1:
        return None
    if to > 1:
        return Convergence.divergence()
    return Convergence.convergence(to)



from typing import List, Union, Tuple, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import sympy as sp
from sympy.plotting.plot import Plot, ContourSeries

from anlis.vector import dotProduct


def contourLines(function: sp.Function,
                 y: sp.Symbol = sp.Symbol('y'),
                 c: Union[float, sp.Symbol] = sp.Symbol('c'),
                 toZero: bool = False,
                 toY: bool = False) -> Union[sp.Eq, List[sp.Eq]]:
    """Returns the contour lines of a function of two variables.
    :param function: function of two variables
    :param y: variable y (if toY is True, y is the variable to solve for)
    :param c: constant c (this variable will be inserted into the function)
    :param toZero: if True, the function is set to zero (c will still be inserted)
    :param toY: if True, the function is set to y (c will still be inserted)
    :return: contour lines of function

    Example:
    >>> from sympy import symbols, sin, cos
    >>> from anlis.multidimensional.contour import contourLines
    >>> x, y, c = symbols('x y c')

    >>> contourLines(x**2 + y**2)
    Eq(x**2 + y**2, c)

    >>> contourLines(x**2 + y**2, y, c)
    Eq(x**2 + y**2, c)

    >>> contourLines(x**2 + y**2, toZero=True)
    Eq(x**2 + y**2 - c, 0)

    >>> contourLines(x**2 + y**2, toY=True)
    [Eq(y, -sqrt(c - x**2)), Eq(y, sqrt(c - x**2))]
    """
    if toZero:
        return sp.Eq(function - c, 0)
    if toY:
        sol = sp.solve(function - c, y)
        return [ sp.Eq(y, s) for s in sol ]
    return sp.Eq(function, c)

def contourPlotSp(function: sp.Function,
                x: sp.Symbol = sp.Symbol('x'),
                y: sp.Symbol = sp.Symbol('y'),
                xmin: int = -10,
                xmax: int = 10,
                ymin: int = -10,
                ymax: int = 10) -> Plot:
    """Returns the contour plot of a function of two variables using SymPy.
    :param function: function of two variables
    :param x: variable x
    :param y: variable y
    :param xmin: minimum value of x
    :param xmax: maximum value of x
    :param ymin: minimum value of y
    :param ymax: maximum value of y
    :return: contour plot of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.contour import contourPlotSp
    >>> x, y = symbols('x y')
    >>> contourPlotSp(x**2 + y**2).show()
    """
    return Plot(ContourSeries(function, (x, xmin, xmax), (y, ymin, ymax)))

def contourPlot(function: sp.Function,
                x: sp.Symbol = sp.Symbol('x'),
                y: sp.Symbol = sp.Symbol('y'),
                xmin: int = -10,
                xmax: int = 10,
                ymin: int = -10,
                ymax: int = 10,
                count: int = 20) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Returns the contour plot of a function of two variables using NumPy.
    :param function: function of two variables
    :param x: variable x
    :param y: variable y
    :param xmin: minimum value of x
    :param xmax: maximum value of x
    :param ymin: minimum value of y
    :param ymax: maximum value of y
    :param count: number of contour lines
    :return: contour plot of function

    Example:
    >>> import matplotlib.pyplot as plt
    >>> from sympy import symbols
    >>> from anlis.multidimensional.contour import contourPlot
    >>> x, y = symbols('x y')
    >>> X, Y, Z = contourPlot(x**2 + y**2)
    >>> plt.contour(X, Y, Z, levels = 10) # levels sets the number of contour lines
    >>> plt.show()
    """
    if count < 2:
        raise ValueError('Count must be greater than 1.')
    if count < 10:
        print('Warning: count is less than 10 - the contour plot may not be smooth.')

    xx = np.linspace(xmin, xmax, count)
    yy = np.linspace(ymin, ymax, count)
    X, Y = np.meshgrid(xx, yy)
    Z = sp.lambdify((x, y), function, 'numpy')(X, Y)
    return X, Y, Z

def contourTangent(gradient: Tuple[sp.Function],
                   x0: Tuple[float],
                   variables: Tuple[sp.Symbol]) -> sp.Function:
    """Returns the tangent to a contour line.
    :param gradient: gradient of the function
    :param x0: point on the function
    :param variables: variables of the tangent
    :return: tangent to the contour line
    """
    assert len(gradient) == len(x0) == len(variables)

    gx0 = [ g.subs({variables[i]: x0[i] for i in range(len(variables))}) for g in gradient ]
    dx = [ variables[i] - x0[i] for i in range(len(variables)) ]
    return dotProduct(tuple(gx0), dx)


class View:
    def __init__(self, elevation: float = 30, azimuth: float = 240) -> None:
        """Creates a 3D view.
        :param elevation: elevation of the view
        :param azimuth: azimuth of the view
        """
        self.elevation = elevation
        self.azimuth = azimuth
    
    @staticmethod
    def top() -> 'View':
        """
        Returns the top view.
        
        y-axis is pointing to the right,
        x-axis is pointing to the bottom
        """
        return View(90, 0)
    
    @staticmethod
    def bottom() -> 'View':
        """
        Returns the bottom view.
        
        y-axis is pointing to the right,
        x-axis is pointing to the top
        """
        return View(270, 0)
    
    @staticmethod
    def front() -> 'View':
        """
        Returns the front view.

        y-axis is pointing to the right,
        z-axis is pointing to the top
        """
        return View(0, 0)

    @staticmethod
    def back() -> 'View':
        """
        Returns the back view.

        y-axis is pointing to the left,
        z-axis is pointing to the top
        """
        return View(0, 180)

    @staticmethod
    def left() -> 'View':
        """
        Returns the left view.

        x-axis is pointing to the right,
        z-axis is pointing to the top
        """
        return View(0, 270)

    @staticmethod
    def right() -> 'View':
        """
        Returns the right view.

        x-axis is pointing to the left,
        z-axis is pointing to the top
        """
        return View(0, 90)

    @staticmethod
    def isometric() -> 'View':
        """Returns the isometric view."""
        return View(30, 240)

def contourPlot3d(function: sp.Function,
                  x: sp.Symbol = sp.Symbol('x'),
                  y: sp.Symbol = sp.Symbol('y'),
                  z: sp.Symbol = sp.Symbol('z'),
                  xrange: Tuple[int] = (-10, 10),
                  yrange: Tuple[int] = (-10, 10),
                  zrange: Optional[Tuple[int]] = None,
                  view: View = View.isometric(),
                  count: int = 100) -> mpl.figure.Figure:
    """Plots the contour plot of a function of two variables using Matplotlib.
    :param function: function of two variables
    :param x: variable x
    :param y: variable y
    :param z: variable z
    :param xrange: range of x
    :param yrange: range of y
    :param zrange: range of z, if None, the range is automatically determined by min and max values
    :param view: view of the plot
    :param count: number of contour lines
    :return: None

    Example:
    >>> import sympy as sp
    >>> from anlis.multidimensional.contour import contourPlot3d, View
    >>> x, y = sp.symbols('x y')
    >>> f = 2*x**2 - 3*x*y + 8*y**2 + x - y
    >>> contourPlot3d(f, x, y)
    >>> contourPlot3d(f, x, y, view = View.top())
    >>> contourPlot3d(f, x, y, xrange = (-5, 5), yrange = (-5, 5), count = 50)
    >>> contourPlot3d(f, x, y, zrange = (-10, 10))
    """
    if count < 2:
        raise ValueError('Count must be greater than 1.')
    if count < 10:
        print('Warning: count is less than 10 - the contour plot may not be smooth.')
    
    xmin, xmax = xrange
    ymin, ymax = yrange

    xx = np.linspace(xmin, xmax, count)
    yy = np.linspace(ymin, ymax, count)

    X, Y = np.meshgrid(xx, yy)
    Z = sp.lambdify((x, y), function, 'numpy')(X, Y)

    if zrange is None:
        zmin, zmax = Z.min(), Z.max()
    else:
        zmin, zmax = zrange

    zz = np.linspace(zmin, zmax, count)

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.set_xlabel(x.name)
    ax.set_ylabel(y.name)
    ax.set_zlabel(z.name)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    ax.set_zlim(zmin, zmax)
    ax.view_init(elev = view.elevation, azim = view.azimuth)
    
    ax.contour(X, Y, Z, zz, colors = ["blue"], linewidths = [0.5])



from typing import Tuple, List, Optional
import sympy as sp

from anlis.vector import unitVector, dotProduct


def partialDerivative(function: sp.Function, variable: sp.Symbol) -> sp.Function:
    """Returns the partial derivative of a function of two variables.
    :param function: function of two variables
    :param variable: variable
    :return: partial derivative of function with respect to variable

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import partialDerivative
    >>> x, y = symbols('x y')
    >>> partialDerivative(x**2 + y**2, x)
    2*x
    >>> partialDerivative(x**2 + y**2, y)
    2*y
    """
    return sp.diff(function, variable).simplify()

def partialDerivatives(function: sp.Function,
                       *variables: sp.Symbol) -> Tuple[sp.Function, ...]:
    """Returns the partial derivatives of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: partial derivatives of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import partialDerivatives
    >>> x, y = symbols('x y')
    >>> partialDerivatives(x**2 + y**2, x, y)
    (2*x, 2*y)
    """
    return tuple(partialDerivative(function, v) for v in variables)

def gradient(function: sp.Function,
             *variables: sp.Symbol) -> Tuple[sp.Function, ...]:
    """Returns the gradient of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: gradient of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import gradient
    >>> x, y = symbols('x y')
    >>> gradient(x**2 + y**2, x, y)
    (2*x, 2*y)
    """
    return tuple(partialDerivative(function, v) for v in variables)

def linearise(function: sp.Function,
              x0: Tuple[float, ...],
              *variables: sp.Symbol) -> sp.Function:
    """Returns the linearisation of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: linearisation of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import linearise
    >>> x, y = symbols('x y')
    >>> linearise(x**2 + y**2, x, y)
    """
    assert len(x0) == len(variables)
    f0 = function.subs({ v: x for v, x in zip(variables, x0) })
    total = f0
    for v, x in zip(variables, x0):
        total += partialDerivative(function, v).subs({ v: x }) * (v - x)
    return total

def jacobiMatrix(functions: List[sp.Function],
                 variables: List[sp.Symbol]) -> sp.Matrix:
    """Returns the Jacobi matrix of a function of any number of variables.
    :param functions: functions of two or more variables
    :param variables: variables
    :return: Jacobi matrix of functions

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import JacobiMatrix
    >>> x, y = symbols('x y')
    >>> JacobiMatrix(x**2 + y**2, x, y)
    Matrix([
    [2*x, 2*y]])
    """
    assert len(functions) == len(variables)
    return sp.Matrix([ gradient(f, *variables) for f in functions ])

def determinant(function: sp.Function,
                variables: List[sp.Symbol]) -> sp.Function:
    """Returns the determinant of the Jacobi matrix of a function of any number of variables.
    :param functions: functions of two or more variables
    :param variables: variables
    :return: determinant of Jacobi matrix of functions

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import determinant
    >>> x, y = symbols('x y')
    >>> determinant(x**2 + y**2, x, y)
    4*x**2
    """
    return partialDerivative2(function, *variables).det()

def newtonRaphson(functions: Tuple[sp.Function],
                  variables: Tuple[sp.Symbol],
                  x0: Tuple[float],
                  n: int = 10) -> Tuple[float, ...]:
    """Returns the solution of a system of equations using Newton-Raphson.
    it stops after n iterations.
    :param functions: functions of two or more variables
    :param variables: variables
    :param x0: initial guess
    :param n: number of iterations
    :return: solution

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import newtonRaphson
    >>> x, y = symbols('x y')
    >>> newtonRaphson([x1**2 + x2**2 - 1, x1 - x2], [x1, x2], [1, 2], 2)
    (3/4, 3/4)
    """
    assert len(functions) == len(variables)
    assert len(x0) == len(variables)
    x = sp.Matrix(x0)
    J = jacobiMatrix(functions, variables)

    for _ in range(n):
        f = sp.Matrix([ f.subs({ v: x for v, x in zip(variables, list(x)) })
                        for f in functions ])
        x = x - J.inv().subs({ v: x for v, x in zip(variables, list(x)) }) * f
    return tuple(x)

def newtonRaphsonWithEpsilon(functions: Tuple[sp.Function],
                             variables: Tuple[sp.Symbol],
                             x0: Tuple[float],
                             epsilon: float = 1e-6,
                             maxIterations: int = 1000) -> Tuple[int, Tuple[float, ...]]:
    """Returns the solution of a system of equations using Newton-Raphson.
    it stops when the error is less than epsilon or when the maximum number of iterations is reached.
    :param functions: functions of two or more variables
    :param variables: variables
    :param x0: initial guess
    :param epsilon: maximum error
    :param maxIterations: maximum number of iterations
    :return: solution

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import newtonRaphsonWithEpsilon
    >>> x, y = symbols('x y')
    >>> newtonRaphsonWithEpsilon([x1**2 + x2**2 - 1, x1 - x2], [x1, x2], [1, 2], 0.0001)
    (2, (0.707106781186547, 0.707106781186547))
    """
    assert len(functions) == len(variables)
    assert len(x0) == len(variables)
    x = sp.Matrix(x0)
    J = jacobiMatrix(functions, variables)

    for i in range(maxIterations):
        f = sp.Matrix([ f.subs({ v: x for v, x in zip(variables, list(x)) })
                        for f in functions ])
        if f.norm() < epsilon:
            break
        x = x - J.inv().subs({ v: x for v, x in zip(variables, list(x)) }) * f
    return i, tuple(x)

def partialCompositeDerivative(z: sp.Function,
                               functions: Tuple[sp.Eq],
                               symbol: sp.Symbol) -> sp.Function:
    """Returns the partial composite derivative of a function of any number of variables.
    :param z: function of two or more variables
    :param functions: equations
    :param symbol: variable
    :return: partial composite derivative of z

    Example:
    given: z = f(x, y) = x**2 * e**y, x = 4u, y = 3u**2 - 2v
    find: dz/du
    >>> from sympy import symbols, Eq
    >>> from anlis.multidimensional.derivative import partialCompositeDerivative
    >>> u, v = symbols('u v')
    >>> x, y = symbols('x y')
    >>> z = x**2 * exp(y)
    >>> partialCompositeDerivative(z, [Eq(x, 4*u), Eq(y, 3*u**2 - 2*v)], u)
    32*u*(3*u**2 + 1)*exp(3*u**2 - 2*v)
    """

    def _findFunction(symbol: sp.Symbol) -> Optional[sp.Function]:
        # if the symbol is on the left side of the equation, return the right side
        for function in functions:
            if function.lhs == symbol:
                return function.rhs
        return None

    # recursively compose a final function
    def _composeFunction(function: sp.Function) -> sp.Function:

        for variable in function.free_symbols:
            f = _findFunction(variable)
            if not f: # just a symbol
                continue
            function = function.subs({ variable: _composeFunction(f) })
        return function


    print(_composeFunction(z))
    return partialDerivative(_composeFunction(z), symbol)


def partialDerivative2(function: sp.Function,
                       *variables: sp.Symbol) -> sp.Matrix:
    """Returns the second partial derivatives of a function of two variables. (fxx, fxy, fyx, fyy)
    :param function: function of two or more variables
    :param variables: variables
    :return: partial derivative matrix

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.derivative import partialDerivative2
    >>> x, y = symbols('x y')
    >>> partialDerivative2(x**2 + y**2, [x, y])
    Matrix([
        [2, 0],
        [0, 2]])
    """
    assert len(variables) == 2

    f1 = partialDerivatives(function, *variables)
    f2 = [ partialDerivatives(f, *variables) for f in f1 ]
    return sp.Matrix(f2)

def partialDerivativeN(f: sp.Function,
                       n: Tuple[sp.Symbol]) -> sp.Function:
    """Returns the partial derivative of a function of two or more variables.
    :param f: function of two or more variables
    :param n: variables in the order of the partial derivative (f_[xx], f_[xy], f_[xyz], ...)
    :return: partial derivative
    """
    g = f
    for i in n:
        g = partialDerivative(g, i)
    return g

def directionalDerivative(f: sp.Function,
                          x0: Tuple[float],
                          e: Tuple[float],
                          *variables: sp.Symbol) -> float:
    """Returns the directional derivative of a function of two or more variables.
    :param f: function of two or more variables
    :param x0: initial point
    :param e: direction vector (NOT the unit vector)
    :return: directional derivative
    """
    assert len(variables) == len(x0)
    assert len(variables) == len(e)
    g = gradient(f, *variables)
    g0 = [ g.subs({ v: x for v, x in zip(variables, x0) }) for g in g ]
    ev = unitVector(*e)
    return dotProduct(g0, ev)



from typing import Tuple
import sympy as sp


def magnitude(*vector: sp.Function) -> sp.Function:
    """Returns the magnitude of a vector.
    :param vector: vector
    :return: magnitude of vector

    Example:
    >>> from sympy import symbols
    >>> from anlis.vectors.vectors import magnitude
    >>> x, y = symbols('x y')
    >>> magnitude(x, y)
    sqrt(x**2 + y**2)
    """
    return sp.sqrt(sum(v**2 for v in vector))

def unitVector(*vector: sp.Function) -> Tuple[sp.Function, ...]:
    """Returns the unit vector of a vector.
    :param vector: vector
    :return: unit vector of vector

    Example:
    >>> from sympy import symbols
    >>> from anlis.vectors.vectors import unitVector
    >>> x, y = symbols('x y')
    >>> unitVector(x, y)
    (x/sqrt(x**2 + y**2), y/sqrt(x**2 + y**2))
    """
    return tuple(v/magnitude(*vector) for v in vector)

def dotProduct(*vectors: Tuple[sp.Function, ...]) -> sp.Function:
    """Returns the dot product of two vectors.
    :param vectors: vectors
    :return: dot product of vectors

    Example:
    >>> from sympy import symbols
    >>> from anlis.vectors.vectors import dotProduct
    >>> x, y = symbols('x y')
    >>> dotProduct((x, y), (x, y))
    x**2 + y**2
    """
    return sum(v1*v2 for v1, v2 in zip(*vectors))



import numpy as np

from anlis.types import Function


def RiemannSum(f: Function,
               a: float,
               b: float,
               n: int,
               upper: bool = False) -> float:
    """
    Riemann sum for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :param upper: if True, use upper Riemann sum
    :return: approximation of integral
    """
    h = (b-a)/n
    x = np.linspace(a, b, n+1)
    y = f(x)
    if upper:
        return h * sum(y[1:])
    return h * sum(y[:-1])

def TrapezoidalRule(f: Function,
                    a: float,
                    b: float,
                    n: int) -> float:
    """
    Trapezoidal rule for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :return: approximation of integral
    """
    h = (b-a)/n
    x = np.linspace(a, b, n+1)
    y = f(x)
    return h/2 * (y[0] + 2*sum(y[1:-1]) + y[-1])

def SimpsonsRule(f: Function,
                 a: float,
                 b: float,
                 n: int) -> float:
    """
    Simpson's rule for approximating the integral of f(x) in [a, b].
    :param f: function f(x)
    :param a: left endpoint of interval
    :param b: right endpoint of interval
    :param n: number of subintervals
    :return: approximation of integral
    """
    h = (b-a)/n
    x = np.linspace(a, b, n+1)
    y = f(x)
    return h/3 * (y[0] + 4*sum(y[1:-1:2]) + 2*sum(y[2:-1:2]) + y[-1])



from typing import List, Dict
import sympy as sp
from anlis.multidimensional.derivative import gradient, determinant


def criticalPoints(function: sp.Function, *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the critical points of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: critical points of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import criticalPoints
    >>> x, y = symbols('x y')
    >>> criticalPoints(x**2 + y**2, x, y)
    [{x: 0, y: 0}]
    """
    return sp.solve(gradient(function, *variables), *variables, dict=True)

def minimas(function: sp.Function, *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the minimas of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: minimas of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import minimas
    >>> x, y = symbols('x y')
    >>> minimas(x**2 + y**2, x, y)
    [{x: 0, y: 0}]
    """
    return [ point
             for point in criticalPoints(function, *variables)
             if determinant(function, list(variables)).subs(point) > 0 ]

def maximas(function: sp.Function, *variables: sp.Symbol) -> List[Dict[sp.Symbol, sp.Number]]:
    """Returns the maximas of a function of any number of variables.
    :param function: function of two or more variables
    :param variables: variables
    :return: maximas of function

    Example:
    >>> from sympy import symbols
    >>> from anlis.multidimensional.criticalPoints import maximas
    >>> x, y = symbols('x y')
    >>> maximas(x**2 + y**2, x, y)
    []
    """
    return [ point
             for point in criticalPoints(function, *variables)
             if determinant(function, list(variables)).subs(point) < 0 ]
